/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Communities System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2001 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.core.impl;

import java.sql.*;
import java.util.*;
import org.apache.log4j.*;
import com.silverwrist.util.StringUtil;
import com.silverwrist.venice.db.*;
import com.silverwrist.venice.core.*;
import com.silverwrist.venice.security.AuditRecord;
import com.silverwrist.venice.security.Capability;
import com.silverwrist.venice.security.DefaultLevels;

class SIGCoreData implements SIGData, SIGDataBackend
{
  /*--------------------------------------------------------------------------------
   * Static data members
   *--------------------------------------------------------------------------------
   */

  private static Category logger = Category.getInstance(SIGCoreData.class.getName());

  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private int refcount = 1;             // object reference count
  private EngineBackend engine;         // pointer to engine back end
  private DataPool datapool;            // pointer to data pool
  private int sigid;                    // ID of this SIG
  private java.util.Date created;       // date/time of database creation 
  private java.util.Date last_access;   // date/time of last access
  private java.util.Date last_update;   // date/time of last update
  private int read_level;               // access level required to read contained data
  private int write_level;              // access level required to write summary data
  private int create_level;             // access level required to create subobjects
  private int delete_level;             // access level required to delete SIG
  private int join_level;               // access level required to become a SIG member
  private int contactid;                // this community's contact ID (geographic info)
  private int host_uid;                 // UID of the community host
  private int category_id;              // category ID for this community
  private boolean hidden_directory;     // hidden from directory listings?
  private boolean hidden_search;        // hidden from keyword searches?
  private boolean members_only;         // only members allowed in this SIG?
  private boolean is_admin;             // is this the administrative SIG?
  private short initial_feature;        // the "initial" feature to display for this SIG
  private String name;                  // the name of this SIG
  private String language;              // the default language code
  private String synopsis;              // the synopsis of the community
  private String rules;                 // the community "rules"
  private String alias;                 // the community alias value
  private boolean public_sig;           // is this a public SIG?
  private BitSet features;              // set of available features
  private Hashtable conf_objects = new Hashtable();    // holder for ConferenceSIGContextImpl objects

  /*--------------------------------------------------------------------------------
   * Constructor
   *--------------------------------------------------------------------------------
   */

  SIGCoreData(EngineBackend engine, DataPool datapool, int sigid) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("new SIGCoreData for SIG " + sigid);
    this.engine = engine;
    this.datapool = datapool;
    this.sigid = sigid;

    Connection conn = null;

    try
    { // get a database connection from this object
      conn = datapool.getConnection();

      // get the SIG basic data from the database
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT * FROM sigs WHERE sigid = ");
      sql.append(sigid).append(';');
      ResultSet rs = stmt.executeQuery(sql.toString());
      if (!(rs.next()))
	throw new DataException("SIG #" + sigid + " does not exist in the database.");

      loadData(rs);  // load the SIG data

      // get the SIG feature set
      sql.setLength(0);
      sql.append("SELECT ftr_code FROM sigftrs WHERE sigid = ").append(sigid).append(';');
      rs = stmt.executeQuery(sql.toString());
      features = new BitSet();
      while (rs.next())
	features.set(rs.getInt("ftr_code"));

    } // end try
    catch (SQLException e)
    { // translate into a nicer DataException
      logger.error("DB error reading SIG data: " + e.getMessage(),e);
      throw new DataException("unable to load SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end constructor

  protected SIGCoreData(EngineBackend engine, DataPool datapool, int sigid, java.util.Date creation,
			String name, String alias, int host_uid, String language, String synopsis,
			String rules, String joinkey, boolean hide_dir, boolean hide_search,
			BitSet features)
  {
    if (logger.isDebugEnabled())
      logger.debug("new SIGCoreData for BRAND NEW SIG " + sigid);
    this.engine = engine;
    this.datapool = datapool;
    this.sigid = sigid;
    this.created = (java.util.Date)(creation.clone());
    this.last_access = this.created;
    this.last_update = this.created;
    this.read_level = DefaultLevels.newSIGRead();
    this.write_level = DefaultLevels.newSIGWrite();
    this.create_level = DefaultLevels.newSIGCreate();
    this.delete_level = DefaultLevels.newSIGDelete();
    this.join_level = DefaultLevels.newSIGJoin();
    this.contactid = -1;
    this.host_uid = host_uid;
    this.category_id = 0;
    this.hidden_directory = hide_dir;
    this.hidden_search = hide_search;
    this.members_only = true;
    this.is_admin = false;
    this.initial_feature = 0;
    this.name = name;
    this.language = language;
    this.synopsis = synopsis;
    this.rules = rules;
    this.alias = alias;
    this.public_sig = StringUtil.isStringEmpty(joinkey);
    this.features = (BitSet)(features.clone());

  } // end constructor

  /*--------------------------------------------------------------------------------
   * finalize() function
   *--------------------------------------------------------------------------------
   */

  protected void finalize()
  {
    engine = null;
    datapool = null;
    created = null;
    last_access = null;
    last_update = null;
    name = null;
    language = null;
    synopsis = null;
    rules = null;
    alias = null;

  } // end finalize

  /*--------------------------------------------------------------------------------
   * Internal functions
   *--------------------------------------------------------------------------------
   */

  private synchronized void loadData(ResultSet rs) throws SQLException
  {
    // "sigid" field is skipped
    created = SQLUtil.getFullDateTime(rs,"createdate");
    last_access = SQLUtil.getFullDateTime(rs,"lastaccess");
    last_update = SQLUtil.getFullDateTime(rs,"lastupdate");
    read_level = rs.getInt("read_lvl");
    write_level = rs.getInt("write_lvl");
    create_level = rs.getInt("create_lvl");
    delete_level = rs.getInt("delete_lvl");
    join_level = rs.getInt("join_lvl");
    contactid = rs.getInt("contactid");
    host_uid = rs.getInt("host_uid");
    category_id = rs.getInt("catid");
    hidden_directory = rs.getBoolean("hide_dir");
    hidden_search = rs.getBoolean("hide_search");
    members_only = rs.getBoolean("membersonly");
    is_admin = rs.getBoolean("is_admin");
    initial_feature = rs.getShort("init_ftr");
    name = rs.getString("signame");
    language = rs.getString("language");
    synopsis = rs.getString("synopsis");
    rules = rs.getString("rules");
    if (!is_admin)
    { // figure out whether we need a "join key" (password)
      String foo = rs.getString("joinkey");
      public_sig = StringUtil.isStringEmpty(foo);

    } // end if
    else  // Admin is NEVER a public SIG!
      public_sig = false;
    alias = rs.getString("alias");

    if (logger.isDebugEnabled())
      logger.debug("Loaded data for SIG \"" + name + "\" (ID " + sigid + ", alias '" + alias
		   + "') successfully");

  } // end loadData

  private synchronized void touchUpdate(Connection conn) throws SQLException
  {
    Statement stmt = conn.createStatement();
    StringBuffer sql = new StringBuffer("UPDATE sigs SET lastupdate = '");
    java.util.Date now = new java.util.Date();
    sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
    stmt.executeUpdate(sql.toString());
    last_update = now;

  } // end touchUpdate

  /*--------------------------------------------------------------------------------
   * Implementations from interface ReferencedData
   *--------------------------------------------------------------------------------
   */

  public int rd_addRef()
  {
    return ++refcount;

  } // end rd_addRef()

  public int rd_release()
  {
    return --refcount;

  } // end rd_release

  public boolean rd_unreferenced()
  {
    return (refcount<=0);

  } // end rd_unreferenced

  /*--------------------------------------------------------------------------------
   * Implementations from interface SIGData
   *--------------------------------------------------------------------------------
   */

  public int getID()
  {
    return sigid;

  } // end getID

  public String getName()
  {
    return name;

  } // end getName

  public String getAlias()
  {
    return alias;

  } // end getAlias

  public boolean isPublicSIG()
  {
    return public_sig;

  } // end isPublicSIG

  public int getCategoryID()
  {
    return category_id;

  } // end getCategoryID

  public String getSynopsis()
  {
    return synopsis;

  } // end getSynopsis

  public int getHostUID()
  {
    return host_uid;

  } // end getHostUID

  public String getLanguage()
  {
    return language;

  } // end getLanguage

  public String getRules()
  {
    return rules;

  } // end getRules

  public int getContactID()
  {
    return contactid;

  } // end getContactID

  public java.util.Date getCreationDate()
  {
    return created;

  } // end getCreationDate

  public java.util.Date getLastAccessDate()
  {
    return last_access;

  } // end getLastAccessDate

  public java.util.Date getLastUpdateDate()
  {
    return last_update;

  } // end getLastUpdateDate

  public void testMembership(int level, boolean is_member) throws AccessError
  {
    if (Capability.exemptFromMembershipRequirement(level))
      return;
    if (members_only && !is_member)
    { // the membership test failed
      logger.error("not a SIG member");
      throw new AccessError("You are not a member of this SIG.");

    } // end if

  } // end testMembership

  public boolean checkMembership(int level, boolean is_member)
  {
    if (Capability.exemptFromMembershipRequirement(level))
      return true;
    return !members_only || is_member;

  } // end checkMembership

  public boolean canReadSIGSubObjects(int level)
  {
    return (level>=read_level);

  } // end canReadSIGSubObjects

  public boolean canModifySIGProfile(int level)
  {
    return (level>=write_level);

  } // end canModifySIGProfile

  public boolean canCreateSIGSubObjects(int level)
  {
    return (level>=create_level);

  } // end canCreateSIGSubObjects

  public boolean canDeleteSIG(int level)
  {
    return (level>=delete_level);

  } // end canDeleteSIG

  public boolean canJoinSIG(int uid, int level)
  {
    if (level<join_level)
      return false;  // rejected on a level basis

    Connection conn = null;  // database connection
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // check to see if the UID is listed in SIGBAN...
      StringBuffer sql = new StringBuffer("SELECT by_uid FROM sigban WHERE sigid = ");
      sql.append(sigid).append(" AND uid = ").append(uid).append(';');
      ResultSet rs = stmt.executeQuery(sql.toString());
      if (rs.next())
	return false;
    
    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.warn("DB error looking up ban info: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return true;

  } // end canJoinSIG

  public synchronized void putContactInfo(UserBackend user, ContactInfo ci) throws DataException
  {
    Connection conn = null;  // database connection
    AuditRecord ar = null;   // audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Stashable obj = (Stashable)ci;

      // save the contact information
      obj.stash(conn);

      if (contactid<0)
      { // contact being established for the first time
	contactid = ci.getContactID();
	if (logger.isDebugEnabled())
	  logger.debug("established new contact ID (" + contactid + ") for this SIG");

      } // end if

      touchUpdate(conn);
      ar = new AuditRecord(AuditRecord.SIG_CONTACT_INFO,user.realUID(),user.userRemoteAddress(),sigid,
			   "contactid=" + contactid);

    } // end try
    catch (ClassCastException cce)
    { // we need to be able to coerce the ContactInfo to a Stashable
      logger.error("ContactInfo is not a Stashable");
      throw new DataException("improper contact information record");

    } // end catch
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG contact info: " + e.getMessage(),e);
      throw new DataException("unable to save user contact data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end putContactInfo

  public BitSet getFeatureSet()
  {
    return (BitSet)(features.clone());

  } // end getFeatureSet

  public synchronized void putFeatureSet(UserBackend user, BitSet set) throws DataException
  {
    Connection conn = null;  // database connection
    AuditRecord ar = null;   // audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      stmt.executeUpdate("LOCK TABLES sigftrs WRITE;");

      try
      { // first, delete all the old entries
	StringBuffer sql = new StringBuffer("DELETE FROM sigftrs WHERE sigid = ");
	sql.append(sigid).append(';');
	stmt.executeUpdate(sql.toString());

	// now, add all the new entries
	sql.setLength(0);
	for (int i=0; i<set.length(); i++)
	{ // test all the bits in the new feature set
	  if (set.get(i))
	  { // append the appropriate element to the SQL statement we're building
	    if (sql.length()==0)
	      sql.append("INSERT INTO sigftrs (sigid, ftr_code) VALUES ");
	    else
	      sql.append(", ");
	    sql.append('(').append(sigid).append(", ").append(i).append(')');

	  } // end if

	} // end for

	if (sql.length()>0)
	{ // add the features to the database
	  sql.append(';');
	  stmt.executeUpdate(sql.toString());

	} // end if

	features = set;

      } // end try
      finally
      { // make sure we unlock the tables before we go
	Statement ulk_stmt = conn.createStatement();
	ulk_stmt.executeUpdate("UNLOCK TABLES;");

      } // end finally

      touchUpdate(conn);
      ar = new AuditRecord(AuditRecord.SIG_FEATURE_SET,user.realUID(),user.userRemoteAddress(),sigid);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG feature set: " + e.getMessage(),e);
      throw new DataException("unable to update feature set: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end putFeatureSet

  public List getSIGFeaturesList(int level)
  {
    return engine.getSIGFeatureSet(features,level,canReadSIGSubObjects(level));

  } // end getSIGFeaturesList

  public String getDefaultApplet()
  {
    StringBuffer buf = new StringBuffer(engine.getAppletForFeature(initial_feature));
    buf.append("?sig=").append(sigid);
    return buf.toString();

  } // end getDefaultApplet

  public synchronized void setName(UserBackend user, String name) throws DataException
  {
    Connection conn = null;  // database connection
    AuditRecord ar = null;   // audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET signame = '");
      sql.append(SQLUtil.encodeString(name)).append("', lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      this.name = name;
      last_update = now;
      ar = new AuditRecord(AuditRecord.SIG_NAME,user.realUID(),user.userRemoteAddress(),sigid,"name=" + name);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setName

  public synchronized void setAlias(UserBackend user, String alias) throws DataException
  {
    Connection conn = null;  // database connection
    AuditRecord ar = null;   // audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET alias = '");
      sql.append(alias).append("', lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      this.alias = alias;
      last_update = now;
      ar = new AuditRecord(AuditRecord.SIG_ALIAS,user.realUID(),user.userRemoteAddress(),sigid,
			   "alias=" + alias);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setAlias

  public synchronized void setCategoryID(UserBackend user, int catid) throws DataException
  {
    Connection conn = null;  // database connection
    AuditRecord ar = null;   // audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET catid = ");
      sql.append(catid).append(", lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      this.category_id = catid;
      last_update = now;
      ar = new AuditRecord(AuditRecord.SIG_CATEGORY,user.realUID(),user.userRemoteAddress(),sigid,
			   "catid=" + catid);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setCategoryID

  public synchronized void setSynopsis(String synopsis) throws DataException
  {
    Connection conn = null;  // database connection
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET synopsis = ");
      sql.append(SQLUtil.encodeStringArg(synopsis)).append(", lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      this.synopsis = synopsis;
      last_update = now;

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setSynopsis

  public synchronized void setLanguage(String language) throws DataException
  {
    Connection conn = null;  // database connection
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET language = '");
      sql.append(SQLUtil.encodeString(language)).append("', lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      this.language = language;
      last_update = now;

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setLanguage

  public synchronized void setRules(String rules) throws DataException
  {
    Connection conn = null;  // database connection
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET rules = ");
      sql.append(SQLUtil.encodeStringArg(rules)).append(", lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      this.rules = rules;
      last_update = now;

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setRules

  public synchronized void touch() throws DataException
  {
    Connection conn = null;  // database connection
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET lastaccess = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      last_access = now;

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end touch

  public boolean getHideDirectory()
  {
    return hidden_directory;

  } // end getHideDirectory

  public boolean getHideSearch()
  {
    return hidden_search;

  } // end getHideSearch

  public synchronized void setHideFlags(UserBackend user, boolean directory, boolean search)
      throws DataException
  {
    Connection conn = null;  // database connection
    AuditRecord ar = null;   // audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET hide_dir = ");
      sql.append(directory ? '1' : '0').append(", hide_search = ").append(search ? '1' : '0');
      sql.append(", lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      hidden_directory = directory;
      hidden_search = search;
      last_update = now;
      ar = new AuditRecord(AuditRecord.SIG_HIDE_INFO,user.realUID(),user.userRemoteAddress(),sigid,
			   "dir=" + directory + ",search=" + search);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setHideFlags

  public boolean getMembersOnly()
  {
    return members_only;

  } // end getMembersOnly

  public synchronized void setMembersOnly(UserBackend user, boolean flag) throws DataException
  {
    Connection conn = null;  // database connection
    AuditRecord ar = null;   // audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET membersonly = ");
      sql.append(flag ? '1' : '0').append(", lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      members_only = flag;
      last_update = now;
      ar = new AuditRecord(AuditRecord.SIG_MEMBERS_ONLY,user.realUID(),user.userRemoteAddress(),sigid,
			   "flag=" + flag);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setMembersOnly

  public short getInitialFeatureIndex()
  {
    return initial_feature;

  } // end getInitialFeatureIndex

  public synchronized void setInitialFeatureIndex(short ndx) throws DataException
  {
    Connection conn = null;  // database connection
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET init_ftr = ");
      sql.append(ndx).append(", lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      initial_feature = ndx;
      last_update = now;

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setInitialFeatureIndex

  public String getJoinKey() throws DataException
  {
    Connection conn = null;  // database connection
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT joinkey FROM sigs WHERE sigid = ");
      sql.append(sigid).append(';');
      ResultSet rs = stmt.executeQuery(sql.toString());
      if (!(rs.next()))
      { // can't find the join key?
	logger.error("SIG " + sigid + "appears to not be there");
	throw new DataException("Unable to find join key for SIG");

      } // end if

      return rs.getString("joinkey");

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error retrieving join key: " + e.getMessage(),e);
      throw new DataException("unable to retrieve join key: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getJoinKey

  public synchronized void setJoinKey(UserBackend user, String key) throws DataException
  {
    Connection conn = null;  // database connection
    AuditRecord ar = null;   // audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET joinkey = ");
      sql.append(SQLUtil.encodeStringArg(key)).append(", lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');
      stmt.executeUpdate(sql.toString());
      public_sig = StringUtil.isStringEmpty(key);
      last_update = now;
      ar = new AuditRecord(AuditRecord.SIG_JOIN_KEY,user.realUID(),user.userRemoteAddress(),sigid);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setJoinKey

  public int getReadLevel()
  {
    return read_level;

  } // end getReadLevel

  public int getWriteLevel()
  {
    return write_level;

  } // end getWriteLevel

  public int getCreateLevel()
  {
    return create_level;

  } // end getCreateLevel

  public int getDeleteLevel()
  {
    return delete_level;

  } // end getDeleteLevel

  public int getJoinLevel()
  {
    return join_level;

  } // end getJoinLevel

  public synchronized void setSecurityLevels(UserBackend user, int read, int write, int create, int delete,
					     int join) throws DataException
  {
    Connection conn = null;  // database connection
    AuditRecord ar = null;   // audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();

      // create the SQL statement
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("UPDATE sigs SET read_lvl = ");
      sql.append(read).append(", write_lvl = ").append(write).append(", create_lvl = ").append(create);
      sql.append(", delete_lvl = ").append(delete).append(", join_lvl = ").append(join);
      sql.append(", lastupdate = '");
      java.util.Date now = new java.util.Date();
      sql.append(SQLUtil.encodeDate(now)).append("' WHERE sigid = ").append(sigid).append(';');

      // execute the database update
      stmt.executeUpdate(sql.toString());

      // save copies of the data to our local variables
      read_level = read;
      write_level = write;
      create_level = create;
      delete_level = delete;
      join_level = join;
      last_update = now;

      // create an audit record reflecting what was done
      ar = new AuditRecord(AuditRecord.SIG_SECURITY,user.realUID(),user.userRemoteAddress(),sigid);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setSecurityLevels

  public boolean isAdminSIG()
  {
    return is_admin;

  } // end isAdminSig

  public synchronized void setMembership(UserBackend user, int uid, int grant_level, boolean locked,
					 boolean hidden) throws DataException
  {
    Connection conn = null;  // database connection
    AuditRecord ar = null;   // successful audit record
      
    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      boolean did_it = false;
      stmt.executeUpdate("LOCK TABLES sigmember WRITE;");

      try
      { // perform the actual manipulation of the SIGMEMBER table...
	StringBuffer sql = new StringBuffer();
	if (grant_level>=0)
	{ // start by seeing if the user already has a record...
	  sql.append("SELECT sigid FROM sigmember WHERE sigid = ").append(sigid).append(" AND uid = ");
	  sql.append(uid).append(';');
	  ResultSet rs = stmt.executeQuery(sql.toString());

	  sql.setLength(0);
	  if (rs.next())
	  { // build an UPDATE statement
	    sql.append("UPDATE sigmember SET granted_lvl = ").append(grant_level).append(", locked = ");
	    sql.append(locked ? '1' : '0').append(", hidden = ").append(hidden ? '1' : '0');
	    sql.append(" WHERE sigid = ").append(sigid).append(" AND uid = ").append(uid).append(';');

	  } // end if
	  else
	  { // build an INSERT statement
	    sql.append("INSERT INTO sigmember (sigid, uid, granted_lvl, locked, hidden) VALUES (");
	    sql.append(sigid).append(", ").append(uid).append(", ").append(grant_level).append(", ");
	    sql.append(locked ? '1' : '0').append(", ").append(hidden ? '1' : '0').append(");");

	  } // end else

	  stmt.executeUpdate(sql.toString());  // update me!
	  did_it = true;

	} // end if
	else
	{ // delete the membership reference (which may not exist, but oh well)
	  sql.append("DELETE FROM sigmember WHERE sigid = ").append(sigid).append(" AND uid = ");
	  sql.append(uid).append(';');
	  did_it = (stmt.executeUpdate(sql.toString())>0);

	} // end else

      } // end try
      finally
      { // make sure the tables are unlocked before we go
	Statement ulk_stmt = conn.createStatement();
	stmt.executeUpdate("UNLOCK TABLES;");

      } // end finally

      if (did_it)
      { // update the SIG data and generate an audit record
	touchUpdate(conn);
	ar = new AuditRecord(AuditRecord.SET_MEMBERSHIP,user.realUID(),user.userRemoteAddress(),sigid,
			     "uid=" + uid,"level=" + grant_level);
      } // end if

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error storing SIG data: " + e.getMessage(),e);
      throw new DataException("unable to save SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end setMembership

  public int getMemberCount(boolean include_hidden) throws DataException
  {
    Connection conn = null;  // database connection
      
    try
    { // get a database connection and create the appropriate SELECT statement
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT COUNT(*) FROM sigmember WHERE sigid = ");
      sql.append(sigid);
      if (!include_hidden)
	sql.append(" AND hidden = 0");
      sql.append(';');

      // execute the query
      ResultSet rs = stmt.executeQuery(sql.toString());
      if (!(rs.next()))
	throw new InternalStateError("getMemberCount query failure - must have ONE row!");

      return rs.getInt(1);

    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error getting SIG data: " + e.getMessage(),e);
      throw new DataException("unable to retrieve SIG data: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getMemberCount

  public boolean isFeaturePresent(String symbol)
  {
    int ndx = engine.getFeatureIndexBySymbol(symbol);
    if (ndx>=0)
      return features.get(ndx);
    else
      return false;

  } // end isFeaturePresent

  public synchronized ConferenceSIGContext getConferenceDataObject(int confid) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("getConferenceDataObject(" + confid + ")...");

    Integer the_confid = new Integer(confid);
    ConferenceSIGContext csc = (ConferenceSIGContext)(conf_objects.get(the_confid));
    if (csc==null)
    { // create a new ConferenceSIGContextImpl object and save it off
      csc = new ConferenceSIGContextImpl(engine,this,datapool,confid);
      conf_objects.put(the_confid,csc);
      if (logger.isDebugEnabled())
	logger.debug("...created new object");

    } // end if
    else
    { // this is an extra reference to the ConferenceSIGContext
      csc.rd_addRef();
      if (logger.isDebugEnabled())
	logger.debug("...using cached object");

    } // end else

    return csc;

  } // end getConferenceDataObject

  public synchronized void detachConferenceDataObject(int confid)
  {
    if (logger.isDebugEnabled())
      logger.debug("detachConferenceDataObject(" + confid + ")...");

    Integer the_confid = new Integer(confid);
    ConferenceSIGContext csc = (ConferenceSIGContext)(conf_objects.get(the_confid));
    if (csc!=null)
    { // remove from the hashtable
      conf_objects.remove(the_confid);
      csc.rd_release();
      if (logger.isDebugEnabled())
	logger.debug("...reference released");

    } // end if

  } // end detachConferenceDataObject

  public ConferenceSIGContext createConference(SIGBackend sig, String name, String alias, String description,
					       boolean pvt, boolean hide_list) throws DataException
  {
    ReturnConfSeq rcs = ConferenceCoreData.createConference(engine,sig,datapool,name,alias,description,
							    pvt,hide_list,host_uid);
    ConferenceData cdata = rcs.getConference();
    if (sig.realUID()!=host_uid)  // make the creating user a conference host too
      cdata.setMembership(sig,sig.realUID(),DefaultLevels.hostConference());

    // Wrap the returned ConferenceData object in a ConferenceSIGContextImpl object and release the
    // extra reference on it.
    ConferenceSIGContextImpl conf = new ConferenceSIGContextImpl(engine,this,datapool,rcs.getSequence(),
								 hide_list,cdata);
    cdata.rd_release();
    rcs = null;

    synchronized (this)
    { // Register this object with our local cache of ConferenceSIGData objects.
      Integer the_confid = new Integer(conf.getConfID());
      if (conf_objects.get(the_confid)!=null)
	throw new InternalStateError("ConfID " + the_confid + " already exists...but it CAN'T HAVE!");

      // throw an extra reference on it and dump it in the object cache
      conf.rd_addRef();
      conf_objects.put(the_confid,conf);

    } // end synchronized block

    return conf;  // pass it up to the next level

  } // end createConference

  public List searchForMembers(int field, int mode, String term, int offset, int count,
			       boolean include_hidden) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("Member search: SIG = " + sigid + ", field = " + field + ", mode = " + mode + ", term '"
		   + term + "', offset = " + offset + ", count = " + count);

    Vector rc = new Vector();    // return from this function
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // create a new SQL statement
      StringBuffer sql = new StringBuffer("SELECT u.uid, u.username, u.description, c.given_name, "
					  + "c.family_name, c.locality, c.region, c.country, m.granted_lvl "
					  + "FROM users u, contacts c, sigmember m WHERE "
					  + "u.contactid = c.contactid AND u.uid = m.uid AND m.sigid = ");
      sql.append(sigid).append(" AND ");
      switch (field)
      { // determine which field to search on
	case SearchMode.FIELD_USER_NAME:
	  sql.append("u.username ");
	  break;

	case SearchMode.FIELD_USER_DESCRIPTION:
	  sql.append("u.description ");
	  break;

	case SearchMode.FIELD_USER_GIVEN_NAME:
	  sql.append("c.given_name ");
	  break;

	case SearchMode.FIELD_USER_FAMILY_NAME:
	  sql.append("c.family_name ");
	  break;

	default:
	  throw new DataException("invalid search field specified");

      } // end switch

      switch (mode)
      { // compose SQL in different ways depending on the search term
	case SearchMode.SEARCH_PREFIX:
	  sql.append("LIKE '").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_SUBSTRING:
	  sql.append("LIKE '%").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_REGEXP:
	  sql.append("REGEXP '").append(SQLUtil.encodeString(term)).append('\'');
	  break;

	default:
	  throw new DataException("invalid search mode specified");

      } // end switch

      if (!include_hidden)
	sql.append(" AND m.hidden = 0");
      sql.append(" AND u.is_anon = 0 ORDER BY u.username LIMIT ").append(offset).append(", ");
      sql.append(count+1).append(';');

      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // launch the search!
      ResultSet rs = stmt.executeQuery(sql.toString());

      while (rs.next())
      { // add all the found users to the list
	UserFoundImpl ufi = new UserFoundImpl(rs.getInt(1),rs.getString(2),rs.getString(3),rs.getString(4),
					      rs.getString(5),rs.getString(6),rs.getString(7),rs.getString(8),
					      rs.getInt(9));
	UserFound tmp = (UserFound)ufi;
	rc.add(tmp);

      } // end while

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading member entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve member information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return new ReadOnlyVector(rc);

  } // end searchForMembers

  public int getSearchMemberCount(int field, int mode, String term, boolean include_hidden)
      throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("User search: SIG = " + sigid + ", field = " + field + ", mode = " + mode + ", term '"
		   + term + "'");

    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT COUNT(*) FROM users u, contacts c, sigmember m WHERE "
					  + "u.contactid = c.contactid AND u.uid = m.uid AND m.sigid = ");
      sql.append(sigid).append(" AND ");

      switch (field)
      { // determine which field to search on
	case SearchMode.FIELD_USER_NAME:
	  sql.append("u.username ");
	  break;

	case SearchMode.FIELD_USER_DESCRIPTION:
	  sql.append("u.description ");
	  break;

	case SearchMode.FIELD_USER_GIVEN_NAME:
	  sql.append("c.given_name ");
	  break;

	case SearchMode.FIELD_USER_FAMILY_NAME:
	  sql.append("c.family_name ");
	  break;

	default:
	  throw new DataException("invalid search field specified");

      } // end switch

      switch (mode)
      { // compose SQL in different ways depending on the search term
	case SearchMode.SEARCH_PREFIX:
	  sql.append("LIKE '").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_SUBSTRING:
	  sql.append("LIKE '%").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_REGEXP:
	  sql.append("REGEXP '").append(SQLUtil.encodeString(term)).append('\'');
	  break;

	default:
	  throw new DataException("invalid search mode specified");

      } // end switch

      if (!include_hidden)
	sql.append(" AND m.hidden = 0");
      sql.append(" AND u.is_anon = 0;");

      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // launch the search!
      ResultSet rs = stmt.executeQuery(sql.toString());

      if (!(rs.next()))
	throw new InternalStateError("getSearchMemberCount failure (MUST have 1 row!)");

      return rs.getInt(1);

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading member entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve member information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getSearchMemberCount

  public List getMemberList(boolean include_hidden) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("Member list: SIG = " + sigid);

    Vector rc = new Vector();    // return from this function
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();

      // create a new SQL statement
      StringBuffer sql = new StringBuffer("SELECT u.uid, u.username, u.description, c.given_name, "
					  + "c.family_name, c.locality, c.region, c.country, m.granted_lvl "
					  + "FROM users u, contacts c, sigmember m WHERE "
					  + "u.contactid = c.contactid AND u.uid = m.uid AND m.sigid = ");
      sql.append(sigid);
      if (!include_hidden)
	sql.append(" AND m.hidden = 0");
      sql.append(" AND u.is_anon = 0 ORDER BY u.username;");

      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // launch the search!
      ResultSet rs = stmt.executeQuery(sql.toString());

      while (rs.next())
      { // add all the found users to the list
	UserFoundImpl ufi = new UserFoundImpl(rs.getInt(1),rs.getString(2),rs.getString(3),rs.getString(4),
					      rs.getString(5),rs.getString(6),rs.getString(7),rs.getString(8),
					      rs.getInt(9));
	UserFound tmp = (UserFound)ufi;
	rc.add(tmp);

      } // end while

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading member entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve member information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return new ReadOnlyVector(rc);

  } // end getMemberList

  /*--------------------------------------------------------------------------------
   * Implementations from interface SIGDataBackend
   *--------------------------------------------------------------------------------
   */

  public int realSIGID()
  {
    return sigid;

  } // end realSIGID

  /*--------------------------------------------------------------------------------
   * External static operations (usable only from within package)
   *--------------------------------------------------------------------------------
   */

  static SIGData createSIG(EngineBackend engine, UserBackend user, DataPool datapool, String name,
			   String alias, int host_uid, String language, String synopsis, String rules,
			   String joinkey, boolean hide_dir, boolean hide_search)
      throws DataException, AccessError
  {
    Connection conn = null;    // database connection
    BitSet def_features = engine.getDefaultFeaturesMask();
    int new_sigid;             // ID of the new SIG
    java.util.Date creation;   // creation date!
    AuditRecord ar = null;     // the audit record
      
    try
    { // get a database connection and create the appropriate SELECT statement
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      stmt.executeUpdate("LOCK TABLES sigs WRITE, sigftrs WRITE, sigmember WRITE;");

      try
      { // first, we need to check if the SIG alias already exists
	StringBuffer sql = new StringBuffer("SELECT sigid FROM sigs WHERE alias = ");
	sql.append(SQLUtil.encodeStringArg(alias)).append(';');
	ResultSet rs = stmt.executeQuery(sql.toString());
	if (rs.next())  // we should get 0 rows back on this one
	  throw new AccessError("A SIG with that alias already exists.");

	// Now, here comes the BIG INSERT to create the SIG!
	sql.setLength(0);
	sql.append("INSERT INTO sigs (createdate, lastaccess, lastupdate, read_lvl, write_lvl, create_lvl, "
		   + "delete_lvl, join_lvl, host_uid, hide_dir, hide_search, signame, language, synopsis, "
		   + "rules, joinkey, alias) VALUES ('");
	creation = new java.util.Date();
	String creation_str = SQLUtil.encodeDate(creation);
	sql.append(creation).append("', '").append(creation).append("', '").append(creation).append("', ");
	sql.append(DefaultLevels.newSIGRead()).append(", ").append(DefaultLevels.newSIGWrite()).append(", ");
	sql.append(DefaultLevels.newSIGCreate()).append(", ").append(DefaultLevels.newSIGDelete());
	sql.append(", ").append(DefaultLevels.newSIGJoin()).append(", ").append(host_uid).append(", ");
	sql.append(hide_dir ? '1' : '0').append(", ").append(hide_search ? '1' : '0').append(", ");
	sql.append(SQLUtil.encodeStringArg(name)).append(", ").append(SQLUtil.encodeStringArg(language));
	sql.append(", ").append(SQLUtil.encodeStringArg(synopsis)).append(", ");
	sql.append(SQLUtil.encodeStringArg(rules)).append(", ").append(SQLUtil.encodeStringArg(joinkey));
	sql.append(", ").append(SQLUtil.encodeStringArg(alias)).append(");");

	if (logger.isDebugEnabled())
	  logger.debug("SQL: " + sql.toString());

	stmt.executeUpdate(sql.toString()); // here we go!

	// Get the SIGID of the new SIG!
	rs = stmt.executeQuery("SELECT LAST_INSERT_ID();");
	if (rs.next())
	  new_sigid = rs.getInt(1);
	else
	{ // we should have gotten a SIGID back!
	  logger.error("readback of new SIGID failed!");
	  throw new InternalStateError("failure on createSIG return query - must have 1 row!");

	} // end else

	if (def_features.length()>0)
	{ // need to build the SIG's defined feature set
	  sql.setLength(0);
	  for (int i=0; i<def_features.length(); i++)
	  { // look for set bits in the feature set
	    if (def_features.get(i))
	    { // build up the SQL statement piece by piece
	      if (sql.length()==0)
		sql.append("INSERT INTO sigftrs (sigid, ftr_code) VALUES ");
	      else
		sql.append(", ");
	      sql.append('(').append(new_sigid).append(", ").append(i).append(')');

	    } // end if

	  } // end for

	  // close out the SQL statement
	  sql.append(';');
	  if (logger.isDebugEnabled())
	    logger.debug("SQL: " + sql.toString());

	  stmt.executeUpdate(sql.toString());  // do it!

	} // end if

	// Lastly, we need to give the "host UID" host privs in the SIG.  The host's membership
	// is "locked" so they can't unjoin and leave the SIG hostless.
	sql.setLength(0);
	sql.append("INSERT INTO sigmember(sigid, uid, granted_lvl, locked) VALUES (").append(new_sigid);
	sql.append(", ").append(host_uid).append(", ").append(DefaultLevels.creatorSIG()).append(", 1);");
	if (logger.isDebugEnabled())
	  logger.debug("SQL: " + sql.toString());
	stmt.executeUpdate(sql.toString());

      } // end try
      finally
      { // unlock the tables before we bail out
	Statement ulk_stmt = conn.createStatement();
	ulk_stmt.executeUpdate("UNLOCK TABLES;");

      } // end finally

      // create an audit record indicating we were successful
      ar = new AuditRecord(AuditRecord.CREATE_SIG,host_uid,user.userRemoteAddress(),new_sigid,"name=" + name,
			   "alias=" + alias);
    
    } // end try
    catch (SQLException e)
    { // database error - this is a DataException
      logger.error("DB error creating SIG: " + e.getMessage(),e);
      throw new DataException("unable to create SIG: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure the connection is released before we go
      try
      { // save off the audit record before we go, though
	if ((ar!=null) && (conn!=null))
	  ar.store(conn);

      } // end try
      catch (SQLException e)
      { // we couldn't store the audit record!
	logger.error("DB error saving audit record: " + e.getMessage(),e);

      } // end catch

      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    // Create the SIGCoreData object representing this SIG and register it with the engine's SIG
    // data object cache.
    SIGCoreData sig = new SIGCoreData(engine,datapool,new_sigid,creation,name,alias,host_uid,language,
				      synopsis,rules,joinkey,hide_dir,hide_search,def_features);
    engine.registerNewSIG(sig);
    return sig;

  } // end createSIG

} // end class SIGCoreData
