/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Communities System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2001 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.core.impl;

import java.sql.*;
import java.util.*;
import org.apache.log4j.*;
import com.silverwrist.venice.db.*;
import com.silverwrist.venice.security.Capability;
import com.silverwrist.venice.security.DefaultLevels;
import com.silverwrist.venice.core.*;

class SIGUserContextImpl implements SIGContext, SIGBackend
{
  /*--------------------------------------------------------------------------------
   * Internal class for caching the SIG name and alias
   *--------------------------------------------------------------------------------
   */

  class SIGSimpleDataCache
  {
    private String name;
    private String alias;

    public SIGSimpleDataCache(String name, String alias)
    {
      this.name = name;
      this.alias = alias;

    } // end constructor

    public String getName()
    {
      return name;

    } // end getName

    public String getAlias()
    {
      return alias;

    } // end getAlias

  } // end class SIGSimpleDataCache

  /*--------------------------------------------------------------------------------
   * Static data members
   *--------------------------------------------------------------------------------
   */

  private static Category logger = Category.getInstance(SIGUserContextImpl.class.getName());

  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private EngineBackend engine;       // pointer to the engine back end
  private UserBackend user;           // pointer to the user back end
  private DataPool datapool;          // pointer to the main data pool
  private int sigid;                  // ID of the underlying SIG
  private int level;                  // access level we have to the SIG
  private boolean is_member;          // are we a member of the SIG?
  private boolean show_admin;         // do we show up as an admin?
  private boolean locked;             // is our membership locked?
  private SIGSimpleDataCache cache;   // cache object for name and alias
  private SIGData sigdata = null;     // the actual SIG data

  /*--------------------------------------------------------------------------------
   * Constructors
   *--------------------------------------------------------------------------------
   */

  protected SIGUserContextImpl(EngineBackend engine, UserBackend user, DataPool datapool, int sigid,
			       int granted_level, boolean locked, String name, String alias)
  {
    if (logger.isDebugEnabled())
      logger.debug("SIGUserContextImpl constructor:member");
    this.engine = engine;
    this.user = user;
    this.datapool = datapool;
    this.sigid = sigid;
    setMemberValues(granted_level,true,locked);
    this.cache = new SIGSimpleDataCache(name,alias);
    
  } // end constructor

  protected SIGUserContextImpl(EngineBackend engine, UserBackend user, DataPool datapool, int sigid,
			       String name, String alias)
  {
    if (logger.isDebugEnabled())
      logger.debug("SIGUserContextImpl constructor:ordinary");
    this.engine = engine;
    this.user = user;
    this.datapool = datapool;
    this.sigid = sigid;
    this.level = user.realBaseLevel();
    this.is_member = false;
    this.show_admin = false;
    this.locked = false;
    this.cache = new SIGSimpleDataCache(name,alias);

  } // end constructor

  SIGUserContextImpl(EngineBackend engine, UserBackend user, DataPool datapool, SIGData sigdata)
  {
    if (logger.isDebugEnabled())
      logger.debug("SIGUserContextImpl constructor:newSIG");
    this.engine = engine;
    this.user = user;
    this.datapool = datapool;
    this.sigid = sigdata.getID();
    this.cache = null;  // no cache required - we have the SIGData
    sigdata.rd_addRef();
    this.sigdata = sigdata;
    setMemberValues(DefaultLevels.creatorSIG(),true,true);

  } // end constructor

  /*--------------------------------------------------------------------------------
   * finalize() function
   *--------------------------------------------------------------------------------
   */

  protected void finalize()
  {
    if (sigdata!=null)
      sigdata.rd_release();
    engine = null;
    user = null;
    datapool = null;
    cache = null;
    sigdata = null;

  } // end finalize

  /*--------------------------------------------------------------------------------
   * Internal functions
   *--------------------------------------------------------------------------------
   */

  private void setMemberValues(int granted_level, boolean member, boolean locked)
  {
    if (logger.isDebugEnabled())
      logger.debug("setMemberValues(" + String.valueOf(granted_level) + ", " + String.valueOf(member)
		   + ", " + String.valueOf(locked));

    if (user.realBaseLevel()>granted_level)
      this.level = user.realBaseLevel();
    else
      this.level = granted_level;
    this.is_member = member;
    this.show_admin = Capability.isCommunityAdmin(granted_level);
    this.locked = locked;

  } // end setMemberValues

  private SIGData getSIGData() throws DataException
  {
    if (sigdata==null)
    { // attempt to load the SIGDataObject
      sigdata = engine.getSIGDataObject(sigid);

      // clear cache when we get the real sigdata
      cache = null;

    } // end if

    return sigdata;

  } // end getSIGData

  private SIGData getSIGDataNE()
  {
    if (sigdata==null)
    { // we need to load the SIGData...
      try
      { // attempt to load the SIGDataObject
	sigdata = engine.getSIGDataObject(sigid);

      } // end try
      catch (DataException e)
      { // there was a data failure loading the SIG
	logger.error("** Got DataException in getSIGDataNE: " + e.getMessage(),e);
	return null;

      } // end catch

      // clear cache when we get the real sigdata
      cache = null;

    } // end if

    return sigdata;

  } // end getSIGData

  private void testConferenceAccess() throws DataException, AccessError
  {
    if (!(getSIGData().isFeaturePresent("CONF")))
    { // the SIG doesn't use conferencing
      logger.error("the SIG doesn't use conferencing");
      throw new AccessError("This SIG does not use the conferencing feature.");

    } // end if

    getSIGData().testMembership(level,is_member);
    if (!(engine.canAccessFeature("CONF",level,getSIGData().canReadSIGSubObjects(level))))
    { // you can't access the conferences!
      logger.error("user not permitted to read confs from this SIG");
      throw new AccessError("You are not permitted to access this SIG's conferences.");

    } // end if

  } // end testConferenceAccess

  private static SIGUserContextImpl getSIGPrivate(EngineBackend engine, UserBackend user, DataPool datapool,
						  Connection conn, int sigid) throws DataException
  {
    try
    { // create the query to find the SIG in the table
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT signame, alias FROM sigs WHERE sigid = ");
      sql.append(sigid).append(';');
      ResultSet rs = stmt.executeQuery(sql.toString());
      if (!(rs.next()))
      { // the SIG entry was not found
	logger.error("SIG " + String.valueOf(sigid) + " not found in database");
	throw new DataException("SIG #" + String.valueOf(sigid) + " was not found in the database.");

      } // end if

      // initialize the object and check membership info
      SIGUserContextImpl sc = new SIGUserContextImpl(engine,user,datapool,sigid,rs.getString("signame"),
						     rs.getString("alias"));
      sc.checkMembership(conn);
      return sc;

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading SIG entry: " + e.getMessage(),e);
      throw new DataException("unable to retrieve SIG information: " + e.getMessage(),e);

    } // end catch

  } // end getSIGPrivate

  /*--------------------------------------------------------------------------------
   * Implementations from class SIGContext
   *--------------------------------------------------------------------------------
   */

  public int getSIGID()
  {
    return sigid;

  } // end getSIGID

  public String getName()
  {
    if (cache!=null)
    { // return cached name if possible
      if (cache.getName()!=null)
	return cache.getName();

    } // end if

    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getName();
    else
      return null;

  } // end getName

  public String getAlias()
  {
    if (cache!=null)
    { // return cached alias if possible
      if (cache.getAlias()!=null)
	return cache.getAlias();

    } // end if

    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getAlias();
    else
      return null;

  } // end getAlias

  public boolean isMember()
  {
    return is_member;

  } // end isMember

  public boolean isAdmin()
  {
    return show_admin;

  } // end isAdmin

  public boolean isPublicSIG()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.isPublicSIG();
    else
      return false;

  } // end isPublicSIG

  public int getCategoryID()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getCategoryID();
    else
      return -1;

  } // end getCategoryID

  public CategoryDescriptor getCategory() throws DataException
  {
    return new CategoryDescriptorImpl(datapool,getSIGData().getCategoryID(),
				      Capability.hideHiddenCategories(user.realBaseLevel()));

  } // end getCategory

  public String getSynopsis()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getSynopsis();
    else
      return null;

  } // end getSynopsis

  public int getHostUID()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getHostUID();
    else
      return -1;

  } // end getHostUID

  public UserProfile getHostProfile() throws DataException
  {
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();

      // load the profile for the user
      return new UserProfileImpl(engine,conn,getSIGData().getHostUID(),
				 Capability.canSeeHiddenContactFields(user.realBaseLevel()));

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      throw new DataException("unable to retrieve host contact information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getHostProfile

  public String getLanguageCode()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getLanguage();
    else
      return null;

  } // end getLanguageCode

  public String getLanguageFullName()
  {
    String code = getLanguageCode();
    if (code!=null)
      return engine.getLanguageNameForCode(code);
    else
      return null;

  } // end getLanguageFullName

  public String getRules()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getRules();
    else
      return null;

  } // end getRules

  public java.util.Date getCreationDate()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getCreationDate();
    else
      return null;

  } // end getCreationDate

  public java.util.Date getLastAccessDate()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getLastAccessDate();
    else
      return null;

  } // end getLastAccessDate

  public java.util.Date getLastUpdateDate()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getLastUpdateDate();
    else
      return null;

  } // end getLastUpdateDate

  public ContactInfo getContactInfo() throws DataException
  {
    int cid = getSIGData().getContactID();
    ContactInfo rc;
    if (cid>=0)
      rc = new ContactInfoImpl(datapool,cid);
    else
      rc = new ContactInfoImpl(getSIGData().getHostUID(),sigid);
    getSIGData().touch();
    return rc;

  } // end getContactInfo

  public void putContactInfo(ContactInfo ci) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (ci.getOwnerSIGID()!=sigid)
    { // this isn't a valid contact info for the SIG
      logger.error("ContactInfo wrong ownership (was " + String.valueOf(ci.getOwnerSIGID())
		   + ", should be " + String.valueOf(sigid) + ")");
      throw new DataException("invalid contact information record");

    } // end if

    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG contact info
      logger.error("user not permitted to modify the SIG's contact info");
      throw new AccessError("You are not permitted to modify the SIG's contact information.");

    } // end if

    getSIGData().putContactInfo(user,ci);

  } // end putContactInfo

  public BitSet getFeatures()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getFeatureSet();
    else
      return null;

  } // end getFeatures

  public void setFeatures(BitSet feature_set, BitSet mask) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's feature set");
      throw new AccessError("You are not permitted to modify the SIG's feature set.");

    } // end if

    // Adjust the "mask"...this function cannot affect the set of "locked" features.
    BitSet real_mask = (BitSet)(mask.clone());
    real_mask.andNot(engine.getLockedFeaturesMask());

    // Figure out which bits are being carried over from the old feature set.
    BitSet new_features = getSIGData().getFeatureSet();
    new_features.andNot(real_mask);

    // Figure out which bits are being set by the new feature set.
    BitSet update_bits = (BitSet)(feature_set.clone());
    update_bits.and(real_mask);

    // Put the features together to result in the final feature set, which we set down in
    // the "back end" class.
    new_features.or(update_bits);
    getSIGData().putFeatureSet(user,new_features);

  } // end setFeatures

  public List getSIGFeaturesList()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getSIGFeaturesList(level);
    else
      return null;

  } // end getSIGFeaturesList

  public String getDefaultApplet()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getDefaultApplet();
    else
      return null;

  } // end getDefaultApplet

  public void setName(String name) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    getSIGData().setName(user,name);

  } // end setName

  public void setAlias(String alias) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    getSIGData().setAlias(user,alias);

  } // end setAlias

  public void setCategoryID(int catid) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    if (getSIGData().isAdminSIG())
    { // the Admin SIG cannot change its security
      logger.error("trying to change category for the ADMIN SIG");
      throw new AccessError("The administrative SIG's category cannot be modified.");

    } // end if

    getSIGData().setCategoryID(user,catid);

  } // end setCategoryID

  public void setCategory(CategoryDescriptor cat) throws DataException, AccessError
  {
    setCategoryID(cat.getCategoryID());

  } // end setCategory

  public void setSynopsis(String synopsis) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    getSIGData().setSynopsis(synopsis);

  } // end setSynopsis

  public void setLanguage(String language) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    getSIGData().setLanguage(language);

  } // end setLanguage

  public void setRules(String rules) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    getSIGData().setRules(rules);

  } // end setRules

  public int getHideMode()
  {
    SIGData sd = getSIGDataNE();
    if (sd==null)
      return HIDE_NONE;
    if (sd.getHideSearch())
      return HIDE_BOTH;
    else if (sd.getHideDirectory())
      return HIDE_DIRECTORY;
    else
      return HIDE_NONE;

  } // end getHideMode

  public void setHideMode(int mode) throws DataException, AccessError
  {
    if ((mode!=HIDE_NONE) && (mode!=HIDE_DIRECTORY) && (mode!=HIDE_BOTH))
    { // the mode is not valid
      logger.error("mode value " + String.valueOf(mode) + " is not valid");
      throw new IllegalArgumentException("invalid hide mode value");

    } // end if

    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    if (getSIGData().isAdminSIG())
    { // the Admin SIG cannot change its security
      logger.error("trying to change security info for the ADMIN SIG");
      throw new AccessError("The administrative SIG's security information cannot be modified.");

    } // end if

    boolean hide_dir = (mode!=HIDE_NONE);
    boolean hide_search = (mode==HIDE_BOTH);
    getSIGData().setHideFlags(user,hide_dir,hide_search);

  } // end setHideMode

  public boolean getMembersOnly()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getMembersOnly();
    else
      return true;

  } // end getMembersOnly

  public void setMembersOnly(boolean flag) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    if (getSIGData().isAdminSIG())
    { // the Admin SIG cannot change its security
      logger.error("trying to change security info for the ADMIN SIG");
      throw new AccessError("The administrative SIG's security information cannot be modified.");

    } // end if

    getSIGData().setMembersOnly(user,flag);

  } // end setMembersOnly

  public short getInitialFeatureIndex()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.getInitialFeatureIndex();
    else
      return 0;

  } // end getInitialFeatureIndex

  public void setInitialFeatureIndex(short ndx) throws DataException, AccessError
  {
    if (!(engine.isValidInitialFeatureIndex(ndx)))
    { // the mode is not valid
      logger.error("feature index value " + String.valueOf(ndx) + " is not valid");
      throw new IllegalArgumentException("invalid initial feature index");

    } // end if

    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    getSIGData().setInitialFeatureIndex(ndx);

  } // end setInitialFeatureIndex

  public String getJoinKey() throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to retrieve join key");
      throw new AccessError("You are not permitted to retrieve the join key for this SIG.");

    } // end if

    return getSIGData().getJoinKey();

  } // end getJoinKey

  public void setJoinKey(String key) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    if (getSIGData().isAdminSIG())
    { // the Admin SIG cannot change its security
      logger.error("trying to change security info for the ADMIN SIG");
      throw new AccessError("The administrative SIG's security information cannot be modified.");

    } // end if

    getSIGData().setJoinKey(user,key);

  } // end setJoinKey

  public int getReadLevel() throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this SIG.");

    } // end if

    return getSIGData().getReadLevel();

  } // end getReadLevel

  public int getWriteLevel() throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this SIG.");

    } // end if

    return getSIGData().getWriteLevel();

  } // end getWriteLevel

  public int getCreateLevel() throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this SIG.");

    } // end if

    return getSIGData().getCreateLevel();

  } // end getCreateLevel

  public int getDeleteLevel() throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this SIG.");

    } // end if

    return getSIGData().getDeleteLevel();

  } // end getDeleteLevel

  public int getJoinLevel() throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to retrieve security info");
      throw new AccessError("You are not permitted to retrieve security information for this SIG.");

    } // end if

    return getSIGData().getJoinLevel();

  } // end getJoinLevel

  public void setSecurityLevels(int read, int write, int create, int delete, int join)
      throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().canModifySIGProfile(level)))
    { // this user can't modify the SIG feature set
      logger.error("user not permitted to modify the SIG's information");
      throw new AccessError("You are not permitted to modify the SIG's information.");

    } // end if

    if (getSIGData().isAdminSIG())
    { // the Admin SIG cannot change its security
      logger.error("trying to change security info for the ADMIN SIG");
      throw new AccessError("The administrative SIG's security information cannot be modified.");

    } // end if

    if (read>write)
    { // correct the write level...
      if (logger.isDebugEnabled())
	logger.debug("read level (" + String.valueOf(read) + ") greater than write level ("
		     + String.valueOf(write) + "); correcting write level");
      write = read;

    } // end if

    if (write>delete)
    { // correct the delete level...
      if (logger.isDebugEnabled())
	logger.debug("write level (" + String.valueOf(write) + ") greater than delete level ("
		     + String.valueOf(delete) + "); correcting delete level");
      delete = write;

    } // end if

    if (create>delete)
    { // correct the create level...
      if (logger.isDebugEnabled())
	logger.debug("create level (" + String.valueOf(create) + ") greater than delete level ("
		     + String.valueOf(delete) + "); correcting create level");
      create = delete;

    } // end if

    getSIGData().setSecurityLevels(user,read,write,create,delete,join);

  } // end setSecurityLevels

  public boolean canAdministerSIG()
  {
    SIGData sd = getSIGDataNE();
    if (sd==null)
      return false;

    if (sd.canDeleteSIG(level))
      return true;
    if (sd.canModifySIGProfile(level))
      return true;
    if (sd.canCreateSIGSubObjects(level))
      return true;

    return false;
    
  } // end canAdministerSIG

  public boolean canModifyProfile()
  {
    SIGData sd = getSIGDataNE();
    if (sd==null)
      return false;

    return sd.canModifySIGProfile(level);

  } // end canModifyProfile

  public boolean isAdminSIG()
  {
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.isAdminSIG();
    else
      return false;

  } // end isAdminSIG

  public void join(String joinkey) throws DataException, AccessError
  {
    if (is_member)
    { // we're already a member!
      logger.info("attempting to join SIG, but already a member");
      return;  // this becomes a no-op

    } // end if

    if (!(getSIGData().canJoinSIG(user.realUID(),level)))
    { // this user can't join up!
      logger.error("user not permitted to join SIG");
      throw new AccessError("You are not permitted to join this SIG.");

    } // end if

    if (!(getSIGData().isPublicSIG()))
    { // this is a private SIG - but admins can join anyway
      if (!(Capability.canJoinPrivateSIGWithoutKey(level)))
      { // we need to test the join key against the one they supply
	String real_key = getSIGData().getJoinKey();
	if (!(real_key.equals(joinkey)))
	{ // the join key was not the one stored in the SIG
	  logger.error("invalid join key specified");
	  throw new AccessError("The join key you specified was incorrect.");

	} // end if

      } // end if
      // else the admin person can join, that's cool

    } // end if
    // else we can join without specifying a key

    // actually set the data in the database
    getSIGData().setMembership(user,user.realUID(),DefaultLevels.memberSIG(),false,false);

    // and update our internal data store
    setMemberValues(DefaultLevels.memberSIG(),true,false);

    // and that's it!  You expected lightning bolts maybe?

  } // end join

  public void unjoin() throws DataException, AccessError
  {
    if (!is_member)
    { // we're already a member!
      logger.info("attempting to unjoin SIG, but not currently a member");
      return;  // this becomes a no-op

    } // end if

    if (locked)
    { // we're locked - we can't possibly unjoin!
      logger.info("membership in SIG is locked, cannot remove");
      throw new AccessError("You are not permitted to unjoin this SIG.");

    } // end if

    // actually set the data in the database
    getSIGData().setMembership(user,user.realUID(),-1,false,false);

    // and update our internal data store
    setMemberValues(-1,false,false);

  } // end unjoin

  public int getMemberCount() throws DataException
  {
    return getSIGData().getMemberCount(Capability.showHiddenSIGMembers(level));

  } // end getMemberCount

  public boolean canUnjoin()
  {
    return is_member && !locked;

  } // end canUnjoin

  public boolean canJoin()
  {
    if (is_member)
      return false;
    SIGData sd = getSIGDataNE();
    if (sd!=null)
      return sd.canJoinSIG(user.realUID(),level);
    else
      return false;
    
  } // end canJoin

  public List getConferences() throws DataException, AccessError
  {
    testConferenceAccess();
    return ConferenceUserContextImpl.getSIGConferences(engine,this,datapool);

  } // end getConferences

  public ConferenceContext getConferenceContext(int confid) throws DataException, AccessError
  {
    testConferenceAccess();
    return ConferenceUserContextImpl.getConference(engine,this,datapool,confid);

  } // end getConferenceContext

  public ConferenceContext getConferenceContext(String alias) throws DataException, AccessError
  {
    testConferenceAccess();
    return ConferenceUserContextImpl.getConference(engine,this,datapool,alias);

  } // end getConferenceContext

  public ConferenceContext createConference(String name, String alias, String description, boolean pvt,
					    boolean hide_list) throws DataException, AccessError
  {
    getSIGData().testMembership(level,is_member);
    if (!(getSIGData().isFeaturePresent("CONF")))
    { // we can't create a conference in this SIG
      logger.error("this SIG does not use conferencing");
      throw new AccessError("You are not permitted to create conferences in this SIG.");

    } // end if

    if (!(getSIGData().canCreateSIGSubObjects(level)))
    { // this user can't create a conference
      logger.error("user not permitted to create a conference");
      throw new AccessError("You are not permitted to create conferences in this SIG.");

    } // end if

    // call down to the SIG core data to create the conference
    ConferenceSIGContext cdata = getSIGData().createConference(this,name,alias,description,pvt,hide_list);

    // wrap the returned object in a conference user context object and release the extra reference
    ConferenceUserContextImpl rc;
    try
    { // need to wrap this 'cos it can throw DataException
      rc = new ConferenceUserContextImpl(engine,this,datapool,cdata);

    } // end try
    finally
    { // make sure to release that extra reference before we go
      cdata.rd_release();

    } // end finally

    return rc;

  } // end createConference

  public boolean canCreateConference()
  {
    SIGData sd = getSIGDataNE();
    if (sd==null)
      return false;
    if (!(sd.checkMembership(level,is_member)))
      return false;
    if (sd.isFeaturePresent("CONF"))
      return sd.canCreateSIGSubObjects(level);
    else
      return false;

  } // end canCreateConference

  public List searchForMembers(int field, int mode, String term, int offset, int count)
      throws DataException
  {
    return getSIGData().searchForMembers(field,mode,term,offset,count,Capability.showHiddenSIGMembers(level));

  } // end searchForMembers

  public int getSearchMemberCount(int field, int mode, String term) throws DataException
  {
    return getSIGData().getSearchMemberCount(field,mode,term,Capability.showHiddenSIGMembers(level));

  } // end getSearchMemberCount

  public List getMemberList() throws DataException
  {
    return getSIGData().getMemberList(Capability.showHiddenSIGMembers(level));

  } // end getMemberList

  public boolean canManageConferences()
  {
    SIGData sd = getSIGDataNE();
    if (sd==null)
      return false;
    if (!(sd.checkMembership(level,is_member)))
      return false;
    if (sd.isFeaturePresent("CONF"))
      return sd.canCreateSIGSubObjects(level);
    else
      return false;

  } // end canManageConferences

  /*--------------------------------------------------------------------------------
   * Implementations from interface UserBackend
   *--------------------------------------------------------------------------------
   */

  public int realUID()
  {
    return user.realUID();

  } // end realUID

  public int realBaseLevel()
  {
    return user.realBaseLevel();

  } // end realBaseLevel

  public String userRemoteAddress()
  {
    return user.userRemoteAddress();

  } // end userRemoteAddress

  public String userDefaultPseud() throws DataException
  {
    return user.userDefaultPseud();

  } // end userDefaultPseud

  public boolean userIsAnonymous()
  {
    return user.userIsAnonymous();

  } // end userIsAnonymous

  /*--------------------------------------------------------------------------------
   * Implementations from interface SIGBackend
   *--------------------------------------------------------------------------------
   */

  public SIGContext selfSIG()
  {
    return this;

  } // end selfsig;

  public int realSIGID()
  {
    return sigid;

  } // end realSIGID

  public boolean userHideHiddenConferences()
  {
    return Capability.hideHiddenConferences(level);

  } // end userHideHiddenConferences

  public int realSIGLevel()
  {
    return level;

  } // end realSIGLevel

  public ConferenceSIGContext getConferenceDataObject(int confid) throws DataException
  {
    return getSIGData().getConferenceDataObject(confid);

  } // end getConferenceDataObject

  public boolean userCanCreateSubobjects()
  {
    SIGData sd = getSIGDataNE();
    if (sd==null)
      return false;
    if (sd.checkMembership(level,is_member))
      return sd.canCreateSIGSubObjects(level);
    else
      return false;

  } // end userCanCreateSubObjects

  public String realSIGAlias()
  {
    return getAlias();

  } // end realSIGAlias

  public void detachConferenceDataObject(int confid) throws DataException
  {
    getSIGData().detachConferenceDataObject(confid);

  } // end detachConferenceDataObject

  public boolean userCanDeleteSubobjects()
  {
    return userCanCreateSubobjects();

  } // end userCanDeleteSubobjects

  /*--------------------------------------------------------------------------------
   * Static operations for use within the implementation package
   *--------------------------------------------------------------------------------
   */

  static List getMemberSIGEntries(EngineBackend engine, UserBackend user, DataPool datapool)
      throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("getMemberSIGEntries for user #" + String.valueOf(user.realUID()));
    Vector rc = new Vector();    // return from this function
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT sm.sigid, sm.granted_lvl, sm.locked, s.signame, s.alias "
					  + "FROM sigmember sm, sigs s WHERE sm.sigid = s.sigid "
					  + "AND sm.uid = ");
      sql.append(user.realUID()).append(" ORDER BY s.signame;");
      ResultSet rs = stmt.executeQuery(sql.toString());
      while (rs.next())
      { // create the user contexts and add them to the return vector
	int the_sigid = rs.getInt(1);
	if (logger.isDebugEnabled())
	  logger.debug("...found SIG #" + String.valueOf(the_sigid));
	SIGContext tmp = new SIGUserContextImpl(engine,user,datapool,the_sigid,rs.getInt(2),rs.getBoolean(3),
						rs.getString(4),rs.getString(5));
	rc.add(tmp);

      } // end while

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading SIG entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve user SIG information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return new ReadOnlyVector(rc);

  } // end getMemberSIGEntries

  static SIGContext getSIGContext(EngineBackend engine, UserBackend user, DataPool datapool, int sigid)
      throws DataException
  {
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();

      // return the SIG we want
      return getSIGPrivate(engine,user,datapool,conn,sigid);

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading SIG entry: " + e.getMessage(),e);
      throw new DataException("unable to retrieve SIG information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getSIGContext

  static SIGContext getSIGContext(EngineBackend engine, UserBackend user, DataPool datapool, String alias)
      throws DataException
  {
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();

      // create the query to find the SIG in the table
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT sigid, signame FROM sigs WHERE alias = '");
      sql.append(SQLUtil.encodeString(alias)).append("';");
      ResultSet rs = stmt.executeQuery(sql.toString());
      if (!(rs.next()))
      { // no SIG entry found...
	logger.error("SIG '" + alias + "' not found in the database");
	throw new DataException("The SIG '" + alias + "' was not found in the database.");

      } // end if

      // initialize the object and check membership info
      SIGUserContextImpl sc = new SIGUserContextImpl(engine,user,datapool,rs.getInt("sigid"),
						     rs.getString("signame"),alias);
      sc.checkMembership(conn);
      return sc;

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading SIG entry: " + e.getMessage(),e);
      throw new DataException("unable to retrieve SIG information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getSIGContext

  static SIGBackend getSIGBackend(EngineBackend engine, UserBackend user, DataPool datapool, Connection conn,
				  int sigid) throws DataException
  {
    return getSIGPrivate(engine,user,datapool,conn,sigid);

  } // end getSIGBackend

  static List searchForSIGs(EngineBackend engine, UserBackend user, DataPool datapool, int field,
			    int mode, String term, int offset, int count) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("SIG search: field = " + String.valueOf(field) + ", mode = " + String.valueOf(mode)
		   + ", term '" + term + "', offset = " + String.valueOf(offset) + ", count = "
		   + String.valueOf(count));

    Vector rc = new Vector();    // return from this function
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT sigid, signame, alias FROM sigs WHERE ");

      switch (field)
      {
	case SearchMode.FIELD_SIG_NAME:
	  sql.append("signame ");
	  break;

	case SearchMode.FIELD_SIG_SYNOPSIS:
	  sql.append("synopsis ");
	  break;

	default:
	  throw new DataException("invalid search field specified");

      } // end switch

      switch (mode)
      { // compose SQL in different ways depending on the search term
	case SearchMode.SEARCH_PREFIX:
	  sql.append("LIKE '").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_SUBSTRING:
	  sql.append("LIKE '%").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_REGEXP:
	  sql.append("REGEXP '").append(SQLUtil.encodeString(term)).append('\'');
	  break;

	default:
	  throw new DataException("invalid search mode specified");

      } // end switch

      if (Capability.hideHiddenSearchSIGs(user.realBaseLevel()))
	sql.append(" AND hide_search = 0");
      sql.append(" ORDER BY signame LIMIT ").append(offset).append(", ").append(count+1).append(';');

      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // launch the search!
      ResultSet rs = stmt.executeQuery(sql.toString());

      while (rs.next())
      { // load all matching SIGs into the recordset
	SIGUserContextImpl sc = new SIGUserContextImpl(engine,user,datapool,rs.getInt("sigid"),
						       rs.getString("signame"),rs.getString("alias"));
	sc.checkMembership(conn);
	SIGContext tmp = sc;
	rc.add(tmp);

      } // end while

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading SIG entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve user SIG information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return new ReadOnlyVector(rc);

  } // end searchForSIGs

  static int getSearchSIGCount(UserBackend user, DataPool datapool, int field, int mode, String term)
      throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("SIG search: field = " + String.valueOf(field) + ", mode = " + String.valueOf(mode)
		   + ", term '" + term + "'");

    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT COUNT(*) FROM sigs WHERE ");

      switch (field)
      {
	case SearchMode.FIELD_SIG_NAME:
	  sql.append("signame ");
	  break;

	case SearchMode.FIELD_SIG_SYNOPSIS:
	  sql.append("synopsis ");
	  break;

	default:
	  throw new DataException("invalid search field specified");

      } // end switch

      switch (mode)
      { // compose SQL in different ways depending on the search term
	case SearchMode.SEARCH_PREFIX:
	  sql.append("LIKE '").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_SUBSTRING:
	  sql.append("LIKE '%").append(SQLUtil.encodeStringWildcards(term)).append("%'");
	  break;

	case SearchMode.SEARCH_REGEXP:
	  sql.append("REGEXP '").append(SQLUtil.encodeString(term)).append('\'');
	  break;

	default:
	  throw new DataException("invalid search mode specified");

      } // end switch

      if (Capability.hideHiddenSearchSIGs(user.realBaseLevel()))
	sql.append(" AND hide_search = 0");
      sql.append(';');

      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // launch the search!
      ResultSet rs = stmt.executeQuery(sql.toString());

      if (!(rs.next()))
	throw new InternalStateError("getSearchSIGCount search failure (MUST have 1 row!)");

      return rs.getInt(1);

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading SIG entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve user SIG information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getSearchSIGCount

  static List getSIGsInCategory(EngineBackend engine, UserBackend user, DataPool datapool, int catid,
				int offset, int count)
      throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("reading SIGs in category " + String.valueOf(catid) + ", offset = "
		   + String.valueOf(offset) + ", count = " + String.valueOf(count));

    Vector rc = new Vector();    // return from this function
    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT sigid, signame, alias FROM sigs WHERE catid = ");
      sql.append(catid);
      if (Capability.hideHiddenDirectorySIGs(user.realBaseLevel()))
	sql.append(" AND hide_dir = 0");
      sql.append(" ORDER BY signame LIMIT ").append(offset).append(", ").append(count+1).append(';');
      
      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // launch the search!
      ResultSet rs = stmt.executeQuery(sql.toString());

      while (rs.next())
      { // load all matching SIGs into the recordset
	SIGUserContextImpl sc = new SIGUserContextImpl(engine,user,datapool,rs.getInt("sigid"),
						       rs.getString("signame"),rs.getString("alias"));
	sc.checkMembership(conn);
	SIGContext tmp = sc;
	rc.add(tmp);

      } // end while

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading SIG entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve user SIG information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

    return new ReadOnlyVector(rc);

  } // end getSIGsInCategory

  static int getNumSIGsInCategory(UserBackend user, DataPool datapool, int catid) throws DataException
  {
    if (logger.isDebugEnabled())
      logger.debug("reading SIGs in category " + String.valueOf(catid));

    Connection conn = null;      // pooled database connection

    try
    { // get a database connection
      conn = datapool.getConnection();
      Statement stmt = conn.createStatement();
      StringBuffer sql = new StringBuffer("SELECT COUNT(*) FROM sigs WHERE catid = ");
      sql.append(catid);
      if (Capability.hideHiddenDirectorySIGs(user.realBaseLevel()))
	sql.append(" AND hide_dir = 0");
      sql.append(';');

      if (logger.isDebugEnabled())
	logger.debug("SQL: " + sql.toString());

      // launch the search!
      ResultSet rs = stmt.executeQuery(sql.toString());

      if (!(rs.next()))
	throw new InternalStateError("getNumSIGsInCategory search failure (MUST have 1 row!)");

      return rs.getInt(1);

    } // end try
    catch (SQLException e)
    { // turn SQLException into data exception
      logger.error("DB error reading SIG entries: " + e.getMessage(),e);
      throw new DataException("unable to retrieve user SIG information: " + e.getMessage(),e);

    } // end catch
    finally
    { // make sure we release the connection before we go
      if (conn!=null)
	datapool.releaseConnection(conn);

    } // end finally

  } // end getNumSIGsInCategory

  /*--------------------------------------------------------------------------------
   * Operations for use within the class only
   *--------------------------------------------------------------------------------
   */

  protected void checkMembership(Connection conn) throws SQLException
  {
    Statement stmt = conn.createStatement();
    StringBuffer sql = new StringBuffer("SELECT granted_lvl, locked FROM sigmember WHERE sigid = ");
    sql.append(sigid).append(" AND uid = ").append(user.realUID()).append(';');
    ResultSet rs = stmt.executeQuery(sql.toString());
    if (rs.next())
    { // we are a member...
      if (logger.isDebugEnabled())
	logger.debug("we are a member of this SIG");
      setMemberValues(rs.getInt("granted_lvl"),true,rs.getBoolean("locked"));

    } // end if
    else
    { // we're NOT a member...
      if (logger.isDebugEnabled())
	logger.debug("we are NOT a member of this SIG");
      setMemberValues(0,false,false);

    } // end else

  } // end checkMembership

} // end class SIGUserContextImpl
