/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Community System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2001 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.security;

import java.util.*;

public class Role implements Comparable, SecLevels
{
  /*--------------------------------------------------------------------------------
   * Private implementation of ReadOnlyVector
   *--------------------------------------------------------------------------------
   */

  static class ReadOnlyVector extends AbstractList
  {
    private Vector my_vec;  // local vector

    ReadOnlyVector(Vector vec)
    {
      my_vec = vec;
      my_vec.trimToSize();

    } // end constructor

    protected void finalize() throws Throwable
    {
      my_vec = null;
      super.finalize();

    } // end finalize

    public Object get(int index)
    {
      return my_vec.elementAt(index);

    } // end get

    public int size()
    {
      return my_vec.size();
      
    } // end size

  } // end class ReadOnlyVector

  /*--------------------------------------------------------------------------------
   * Static data members
   *--------------------------------------------------------------------------------
   */

  private static Role not_in_list = null;
  private static Role no_access = null;
  private static Role unrestricted_user = null;
  private static Vector global_low = null;
  private static Vector global_high = null;
  private static Vector sig_low = null;
  private static Vector sig_high = null;
  private static Vector conf_low = null;
  private static Vector conf_high = null;

  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private int level;
  private String name;

  /*--------------------------------------------------------------------------------
   * Constructor
   *--------------------------------------------------------------------------------
   */

  protected Role(int level, String name)
  {
    this.level = level;
    this.name = name;

  } // end constructor

  /*--------------------------------------------------------------------------------
   * Internal functions
   *--------------------------------------------------------------------------------
   */

  private static void initAllSets()
  {
    if (not_in_list==null)
      not_in_list = new Role(0,"(not in list)");

    if (no_access==null)
      no_access = new Role(NO_ACCESS,"No Access");

    if (unrestricted_user==null)
      unrestricted_user = new Role(UNRESTRICTED_USER,"'Unrestricted' User");

    if (global_low==null)
    { // initialize the "global lowband" vector
      global_low = new Vector(3);
      global_low.addElement(new Role(GLOBAL_ANONYMOUS,"Anonymous User"));
      global_low.addElement(new Role(GLOBAL_UNVERIFIED,"Unauthenticated User"));
      global_low.addElement(new Role(GLOBAL_NORMAL,"Normal User"));
      global_low.trimToSize();

    } // end if

    if (global_high==null)
    { // initialize the "global highband" vector
      global_high = new Vector(3);
      global_high.addElement(new Role(GLOBAL_ANYADMIN,"Any System Administrator"));
      global_high.addElement(new Role(GLOBAL_PFY,"System Assistant Administrator"));
      global_high.addElement(new Role(GLOBAL_BOFH,"Global System Administrator"));
      global_high.trimToSize();

    } // end if

    if (sig_low==null)
    { // initialize the "SIG lowband" vector
      sig_low = new Vector(1);
      sig_low.addElement(new Role(SIG_MEMBER,"SIG Member"));
      sig_low.trimToSize();

    } // end if

    if (sig_high==null)
    { // initialize the "SIG highband" vector
      sig_high = new Vector(3);
      sig_high.addElement(new Role(SIG_ANYADMIN,"Any SIG Administrator"));
      sig_high.addElement(new Role(SIG_COHOST,"SIG Co-Host"));
      sig_high.addElement(new Role(SIG_HOST,"SIG Host"));
      sig_high.trimToSize();

    } // end if

    if (conf_low==null)
    { // initialize the "conference lowband" vector
      conf_low = new Vector(1);
      conf_low.addElement(new Role(CONFERENCE_MEMBER,"Conference Member"));
      conf_low.trimToSize();

    } // end if

    if (conf_high==null)
    { // initialize the "conference highband" vector
      conf_high = new Vector(2);
      conf_high.addElement(new Role(CONFERENCE_ANYADMIN,"Any Conference Administrator"));
      conf_high.addElement(new Role(CONFERENCE_HOST,"Conference Host"));
      conf_high.trimToSize();

    } // end if

  } // end initAllSets

  /*--------------------------------------------------------------------------------
   * External operations
   *--------------------------------------------------------------------------------
   */

  public int getLevel()
  {
    return level;

  } // end getLevel

  public String getName()
  {
    return name;

  } // end getName

  public String toString()
  {
    StringBuffer buf = new StringBuffer(name);
    buf.append('[').append(level).append(']');
    return buf.toString();

  } // end toString

  public boolean equals(Object obj)
  {
    if (obj==null)
      return false;
    else if (obj==this)
      return true;
    if (obj instanceof Role)
    { // compare levels
      Role other = (Role)obj;
      return (level==other.getLevel());

    } // end if

    return obj.toString().equals(toString());

  } // end equals

  public int hashCode()
  {
    return level;

  } // end hashCode

  public int compareTo(Object obj)
  {
    if (!(obj instanceof Role))
      throw new ClassCastException("comparing non-Role to Role");
    Role other = (Role)obj;
    return level - other.getLevel();

  } // end compareTo

  /*--------------------------------------------------------------------------------
   * External static operations which generate lists of roles
   *--------------------------------------------------------------------------------
   */

  public static List getSIGReadList()
  {
    initAllSets();
    Vector rc = new Vector();
    rc.addAll(global_low);
    rc.addAll(sig_low);
    rc.add(unrestricted_user);
    rc.addAll(sig_high);
    rc.add(global_high.firstElement());
    return new ReadOnlyVector(rc);

  } // end getSIGReadList

  public static List getSIGWriteList()
  {
    initAllSets();
    Vector rc = new Vector();
    rc.addAll(sig_high);
    rc.addAll(global_high);
    return new ReadOnlyVector(rc);

  } // end getSIGWriteList

  public static List getSIGCreateList()
  {
    initAllSets();
    Vector rc = new Vector();
    rc.add(global_low.lastElement());
    rc.addAll(sig_low);
    rc.add(unrestricted_user);
    rc.addAll(sig_high);
    rc.add(global_high.firstElement());
    return new ReadOnlyVector(rc);

  } // end getSIGCreateList

  public static List getSIGDeleteList()
  { 
    initAllSets();
    Vector rc = new Vector();
    rc.addAll(sig_high);
    rc.addAll(global_high);
    rc.add(no_access);
    return new ReadOnlyVector(rc);

  } // end getSIGDeleteList

  public static List getSIGJoinList()
  {
    initAllSets();
    return new ReadOnlyVector(global_low);

  } // end getSIGJoinList

  public static List getConferenceReadList()
  {
    initAllSets();
    Vector rc = new Vector();
    rc.addAll(global_low);
    rc.addAll(sig_low);
    rc.addAll(conf_low);
    rc.add(unrestricted_user);
    return new ReadOnlyVector(rc);

  } // end getConferenceReadList

  public static List getConferencePostList()
  {
    initAllSets();
    Vector rc = new Vector();
    rc.addAll(global_low);
    rc.addAll(sig_low);
    rc.addAll(conf_low);
    rc.add(unrestricted_user);
    rc.addAll(conf_high);
    return new ReadOnlyVector(rc);

  } // return getConferencePostList

  public static List getConferenceCreateList()
  {
    return getConferencePostList();

  } // end getConferenceChangeList

  public static List getConferenceHideList()
  {
    initAllSets();
    Vector rc = new Vector();
    rc.addAll(conf_high);
    rc.addAll(sig_high);
    rc.add(global_high.firstElement());
    return new ReadOnlyVector(rc);

  } // end getConferenceHideList

  public static List getConferenceNukeList()
  {
    return getConferenceHideList();

  } // end getConferenceNukeList

  public static List getConferenceChangeList()
  {
    return getConferenceHideList();

  } // end getConferenceChangeList

  public static List getConferenceDeleteList()
  {
    initAllSets();
    Vector rc = new Vector();
    rc.addAll(sig_high);
    rc.addAll(global_high);
    rc.add(no_access);
    return new ReadOnlyVector(rc);

  } // end getConferenceDeleteList

  public static List getConferenceMemberLevelChoices()
  {
    initAllSets();
    Vector rc = new Vector();
    rc.add(not_in_list);
    rc.addAll(global_low);
    rc.addAll(sig_low);
    rc.addAll(conf_low);
    rc.add(unrestricted_user);
    rc.add(conf_high.lastElement());
    return new ReadOnlyVector(rc);

  } // end getConferenceMemberLevelChoices

} // end class Role
